//==============================================================================
//        * Letsch Informatik *       www.LetsInfo.ch       CH-8636 Wald
//          Beratung, Ausbildung und Realisation in Software-Engineering
//==============================================================================
// Project   : C++-Kurs @ Sultex Limited
// Title     : bung 1: Pointer, Referenzen und Memory-Layout: Loesung
// Author    : Thomas Letsch
// Tab-Width : 2
/*///===========================================================================
* Description:
Programm um das Memory-Layout auf einer bestimmten Plattform und Umgebung
zu untersuchen.
Achtung: Keine Optimierungen beim Kompilieren! 
Sonst ist das Memory-Layout fast nicht nachvollziehbar.
* History    :
01.12.02: Initial Version.
03.12.02: Fomatierung angepasst.
12.12.03: ANSI-Header/using namespace std.
26.09.05: Pointer/Referenzen hinzugefgt.
/*///===========================================================================
//       1         2         3         4         5         6         7         8
//345678901234567890123456789012345678901234567890123456789012345678901234567890
//==============================================================================

#include <iostream>
#include <iomanip>

using namespace std;

// Function-Prototyps:

// Aufgabe 1:
void funcPtr(string **pStr);
void funcRef(string *&pStr);

// Aufgabe 2:
int func1();
int func2();
int fac(int pN);

// dumpMemory():
// Dumps Memory to stdout.
// - pPtr:        Start-Adress from where to dump.
// - pIdentStr:   Identify-String which is written first.
// - pNrOfBytes:  Number of Bytes to be dumped.
void dumpMemory(const void* pPtr, const char* pIdentStr = 0, int pNrOfBytes = 4);

// global Definitions:
int        globalInt        = 1;
const int  globalConstInt   = 2;

int main() {
	
	cout << "Aufgabe 1:" << endl;
	
	string* strPtr = 0;
	string** strPtrPtr = &strPtr;
	funcPtr(strPtrPtr);
	cout << "str = " << **strPtrPtr << endl;
	strPtr = 0;
 	funcRef(strPtr);
	cout << "str = " << *strPtr << endl;
	
	
	cout << "\nAufgabe 2:" << endl;
	
  int mainInt1 = 2147483647;     // 2^31 - 1
  int mainInt2 = -2;
  int* ip = new int;
  *ip = -1;

  dumpMemory(&globalInt,       "globalInt",        4 );
  dumpMemory(&globalConstInt,  "globalConstInt",   4 );
  dumpMemory(&mainInt1,        "mainInt1",         sizeof(mainInt1) );
  dumpMemory(&mainInt2,        "mainInt2",         sizeof(mainInt2) );
  dumpMemory(&mainInt2,        "mainInt2 - 8",     8 );
  dumpMemory(&ip,              "ip",               sizeof(ip) );
  dumpMemory(ip,               "*ip",              sizeof(*ip) );
  dumpMemory((void*)main,      "main()");
  dumpMemory((void*)func1,     "func1()");
  dumpMemory((void*)func2,     "func2()");
  dumpMemory((void*)fac,       "fac()");

  int ret = 0;
  ret = fac(5);
  cout << "fac(5) : " << ret << endl;

} // main()




// Implementation Aufgabe 1:

void funcPtr(string** pStr) {
	*pStr = new string("funcPtr()");
}


void funcRef(string*& pStr) {
	pStr = new string("funcRef()");
}




// Implementation Aufgabe 2:

int func1() {
  return 1;
}


int func2() {
  return 2;
}


int fac(int pN) {
  int newArg = pN - 1;   // Just for Demonstration of local Variable in Functions.
  if (pN > 1) {
    return pN * fac(newArg);
  }
  else {
    dumpMemory(&newArg, "fac(5):newArg", 244); // Let's see the Stack right now !
    return 1;                                  // With 244 Bytes we see enough.
  }
}


void dumpMemory(const void* pPtr, const char* pIdentStr, int pNrOfBytes) {
  if (pIdentStr != 0) {
    cout << pIdentStr << ":" << endl;
  }
  unsigned char* cp = (unsigned char*)pPtr;
  int words = (pNrOfBytes / 4) + (((pNrOfBytes % 4) > 0) ? 1 : 0);
  int byteArr[] = {-1, -1, -1, -1};
  for (int word = 0; word < words; word++) {      // for all Words
    cout << setw(13) << setfill(' ') << static_cast<void*>(cp) << " : ";
    for (int byte = 0; byte < 4; byte++) { // for all Bytes in Word
      // The Bit-Mask: 128dec, left-most bit set to 1:
      unsigned char mask = 0x80;
      for (int bitpos = 0; bitpos < 8; bitpos++) { // for all 8 Bit-Positions
        if (mask & *cp)                            // Mask BIT-AND actual Byte
          cout << "1";
        else
          cout << "0";
        mask = mask >> 1;
      }
      byteArr[byte] = *cp;                         // save Value for Hex print
      cout << " ";
      if ((word * 4) + (byte+1) == pNrOfBytes) {
        break;
      }
      cp++;                                        // Increment Pointer by one Byte
    } // for (int byte = 0; byte < 4; byte++)
    // Append now Hex-Values:
    cout << "   ";
    for (int byte = 0; byte < 4; byte++) {
      if (byteArr[byte] != -1) {
        cout << " " << hex << setw(2) << setfill('0') << byteArr[byte];
      }
    }
    for (int byte = 0; byte < 4; byte++) {
      byteArr[byte] = -1;
    }
    cout << endl;
  } // for (int word = 0; word <= words; word++)
  cout << dec << endl;
} // dumpMemory()

/* Session-Log:

PC/Win200 mit Cygwin/Gnu:
========================

$ uname -a
CYGWIN_NT-5.0 DONNA 1.3.12(0.54/3/2) 2002-07-06 02:16 i686 unknown
$
$ g++ MemoryLayout.cpp -o memlayout.exe
$
$ memlayout.exe
Aufgabe 1:
str = funcPtr()
str = funcRef()

Aufgabe 2:
globalInt:
0x413010 : 00000001 00000000 00000000 00000000  01 00 00 00

globalConstInt:
0x4015a4 : 00000010 00000000 00000000 00000000  02 00 00 00

mainInt1:
0x22fedc : 11111111 11111111 11111111 01111111  ff ff ff 7f

mainInt2:
0x22fed8 : 11111110 11111111 11111111 11111111  fe ff ff ff

mainInt2 - 8:
0x22fed8 : 11111110 11111111 11111111 11111111  fe ff ff ff
0x22fedc : 11111111 11111111 11111111 01111111  ff ff ff 7f

ip:
0x22fed4 : 01000000 00000011 00000001 00001010  40 03 01 0a

*ip:
0xa010340 : 11111111 11111111 11111111 11111111  ff ff ff ff

main():
0x4010a8 : 01010101 10001001 11100101 10000011  55 89 e5 83

func1():
0x401230 : 01010101 10001001 11100101 10111000  55 89 e5 b8

func2():
0x401240 : 01010101 10001001 11100101 10111000  55 89 e5 b8

fac():
0x401260 : 01010101 10001001 11100101 10000011  55 89 e5 83

fac(5):newArg:
0x22fdec : 00000000 00000000 00000000 00000000  00 00 00 00 | newArg : 0          |
0x22fdf0 : 00100000 11111110 00100010 00000000  20 fe 22 00 |                     |
0x22fdf4 : 01110111 00010010 01000000 00000000  77 12 40 00 | PC: fac(): 0x401260 |
0x22fdf8 : 00000001 00000000 00000000 00000000  01 00 00 00 | pN : 1              |
0x22fdfc : 01000100 00110000 01000001 00000000  44 30 41 00 |                     |
0x22fe00 : 00110000 11111110 00100010 00000000  30 fe 22 00 |                     |
0x22fe04 : 01010100 00010011 01000001 00000000  54 13 41 00 |                     |
0x22fe08 : 01000000 00110000 01000001 00000000  40 30 41 00 |                     |
0x22fe0c : 00000001 00000000 00000000 00000000  01 00 00 00 |                     |
0x22fe10 : 10000011 00000000 00000000 00000000  83 00 00 00 +- fac(1) ------------+
0x22fe14 : 00000001 00000000 00000000 00000000  01 00 00 00 |                     |
0x22fe18 : 00000001 00000000 00000000 00000000  01 00 00 00 |                     |
0x22fe1c : 00000001 00000000 00000000 00000000  01 00 00 00 | newArg : 1          |
0x22fe20 : 01010000 11111110 00100010 00000000  50 fe 22 00 |                     |
0x22fe24 : 01110111 00010010 01000000 00000000  77 12 40 00 | PC: fac(): 0x401260 |
0x22fe28 : 00000010 00000000 00000000 00000000  02 00 00 00 | pN : 2              |
0x22fe2c : 01000000 00110000 01000001 00000000  40 30 41 00 |                     |
0x22fe30 : 10110000 11111110 00100010 00000000  b0 fe 22 00 |                     |
0x22fe34 : 10010001 00010101 01000000 00000000  91 15 40 00 |                     |
0x22fe38 : 01000000 00110000 01000001 00000000  40 30 41 00 |                     |
0x22fe3c : 10110100 00101110 01000000 00000000  b4 2e 40 00 |                     |
0x22fe40 : 00000000 00000000 00000000 00000000  00 00 00 00 +- fac(2) ------------+
0x22fe44 : 00000000 00000000 00000000 00000000  00 00 00 00 |                     |
0x22fe48 : 00000000 00000000 00000000 00000000  00 00 00 00 |                     |
0x22fe4c : 00000010 00000000 00000000 00000000  02 00 00 00 | newArg : 2          |
0x22fe50 : 10000000 11111110 00100010 00000000  80 fe 22 00 |                     |
0x22fe54 : 01110111 00010010 01000000 00000000  77 12 40 00 | PC: fac(): 0x401260 |
0x22fe58 : 00000011 00000000 00000000 00000000  03 00 00 00 | pN : 3              |
0x22fe5c : 00000000 00000000 00000000 00000000  00 00 00 00 |                     |
0x22fe60 : 00000000 00000000 00000000 00000000  00 00 00 00 |                     |
0x22fe64 : 00000000 00000000 00000000 00000000  00 00 00 00 |                     |
0x22fe68 : 00011000 00110111 01000001 00000000  18 37 41 00 |                     |
0x22fe6c : 00000100 00000000 00000000 00000000  04 00 00 00 |                     |
0x22fe70 : 10101000 00010110 01000000 00000000  a8 16 40 00 +- fac(3) ------------+
0x22fe74 : 00000010 00000000 00000000 00000000  02 00 00 00 |                     |
0x22fe78 : 01111100 00010110 01000000 00000000  7c 16 40 00 |                     |
0x22fe7c : 00000011 00000000 00000000 00000000  03 00 00 00 | newArg : 3          |
0x22fe80 : 10110000 11111110 00100010 00000000  b0 fe 22 00 |                     |
0x22fe84 : 01110111 00010010 01000000 00000000  77 12 40 00 | PC: fac(): 0x401260 |
0x22fe88 : 00000100 00000000 00000000 00000000  04 00 00 00 | pN : 4              |
0x22fe8c : 00000001 00000000 00000000 00000000  01 00 00 00 |                     |
0x22fe90 : 11111111 11111111 11111111 11111111  ff ff ff ff |                     |
0x22fe94 : 11111111 11111111 11111111 11111111  ff ff ff ff |                     |
0x22fe98 : 11111111 11111111 11111111 11111111  ff ff ff ff |                     |
0x22fe9c : 11111111 11111111 11111111 11111111  ff ff ff ff |                     |
0x22fea0 : 10011000 11111101 00100010 00000000  98 fd 22 00 +- fac(4) ------------+
0x22fea4 : 11000100 00000000 00000000 00000000  c4 00 00 00 |                     |
0x22fea8 : 00000001 00000000 00000000 00000000  01 00 00 00 |                     |
0x22feac : 00000100 00000000 00000000 00000000  04 00 00 00 | newArg : 4          |
0x22feb0 : 11100000 11111110 00100010 00000000  e0 fe 22 00 |                     |
0x22feb4 : 11100101 00010001 01000000 00000000  e5 11 40 00 | PC: main(): 0x4010a8|
0x22feb8 : 00000101 00000000 00000000 00000000  05 00 00 00 | pN : 5              |
0x22febc : 10010110 00010000 01000000 00000000  96 10 40 00 |                     |
0x22fec0 : 00000100 00000000 00000000 00000000  04 00 00 00 |                     |
0x22fec4 : 10110011 00010000 01000000 00000000  b3 10 40 00 |                     |
0x22fec8 : 10110100 00010011 01101000 01100001  b4 13 68 61 |                     |
0x22fecc : 00101111 00000000 00000000 00000000  2f 00 00 00 |                     |
0x22fed0 : 00000000 00000000 00000000 00000000  00 00 00 00 + - fac(5): `return` -+
0x22fed4 : 01000000 00000011 00000001 00001010  40 03 01 0a ip
0x22fed8 : 11111110 11111111 11111111 11111111  fe ff ff ff mainInt2
0x22fedc : 11111111 11111111 11111111 01111111  ff ff ff 7f mainInt1

fac(5) : 120
$

Hinweis: Beschreibungen bei fac(5) sind nachtrglich hinzugefgt.

Memory-Layout:
=============





Bereich                                  Adresse      Datenelement-Name
                                 Richtung
------------------------------------+----------------------------------
                                         0x000000

                                    ^
                                    ^
                                    ^
                                    ^
                                    ^
                                    ^    0x0022fed4   ip
                                    ^    0x0022fed8   mainInt2
                                    ^    0x0022fedc   mainInt1
                                    ^
Stack: -----------------------------+--------------------------------

Code-Segment: ----------------------+ -------------------------------
                                    v
                                    v    0x004010a8   main()
                                    v    0x00401230   func1()
                                    v    0x00401240   func2()
                                    v    0x00401260   fac()
                                    v
                                    v
Daten-Segment: Konstanten: ---------+ -------------------------------
                                    v
                                    v    0x004015a4   globalConstInt
                                    v
                                    v
Daten-Segment: globale Variablen: --+--------------------------------
                                    v
                                    v    0x00413010   globalInt
                                    v
                                    v
Dynamischer Speicher / Heap --------+--------------------------------
                                    v
                                    v    0x0a010340   ip
                                    v
                                    v
                                    v
                                    v
                                    v
                                    v

                                         0xffffff
------------------------------------+--------------------------------




Richtung:
^: Adressen sind abnehmend
v: Adressen sind zunehmend

*/
